# MiloDB by FrozenWolf is marked with CC0 1.0.
# To view a copy of this license, visit http://creativecommons.org/publicdomain/zero/1.0
import re
from dataclasses import dataclass
from typing import Any
from milodb_client.updater.manifest.common_manifest import IPreRelease, IVersionNumber, PreReleaseLabel

class VersionNumberError(Exception):
    pass

class PreRelease(IPreRelease):
    def __init__(self, label: PreReleaseLabel, number: int) -> None:
        self._label: PreReleaseLabel = label
        self._number: int = number

    @property
    def label(self) -> PreReleaseLabel:
        return self._label

    @property
    def number(self) -> int:
        return self._number

    @staticmethod
    def compare(left: IPreRelease, right: IPreRelease) -> int: # pylint: disable=too-many-return-statements
        if left is right:
            return 0
        if left.label.value < right.label.value:
            return -1
        if left.label.value > right.label.value:
            return 1
        if left.number < right.number:
            return -1
        if left.number > right.number:
            return 1
        return 0

    def __str__(self) -> str:
        return f'{self.label.value}.{self.number}'

    def __hash__(self) -> int:
        return hash((self.label.value, self.number))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, IPreRelease):
            return PreRelease.compare(self, other) == 0
        return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, IPreRelease):
            return PreRelease.compare(self, other) != 0
        return False

    def __lt__(self, other: IPreRelease) -> bool:
        return PreRelease.compare(self, other) < 0

    def __le__(self, other: IPreRelease) -> bool:
        return PreRelease.compare(self, other) <= 0

    def __gt__(self, other: IPreRelease) -> bool:
        return PreRelease.compare(self, other) > 0

    def __ge__(self, other: IPreRelease) -> bool:
        return PreRelease.compare(self, other) >= 0

@dataclass
class VersionNumber(IVersionNumber):
    def __init__(self, major: int, minor: int, patch: int, pre_release: PreRelease | None = None) -> None:
        self._major: int = major
        self._minor: int = minor
        self._patch: int = patch
        self._pre_release: PreRelease | None = pre_release

    @property
    def major(self) -> int:
        return self._major
    @property
    def minor(self) -> int:
        return self._minor
    @property
    def patch(self) -> int:
        return self._patch
    @property
    def pre_release(self) -> PreRelease | None:
        return self._pre_release

    @staticmethod
    def parse(text: str) -> IVersionNumber:
        """
        ### Raises:
        - VersionNumberError
        """
        match: re.Match[str] | None = _VERSION_NUMBER_PATTERN.fullmatch(text)

        try:
            if not match:
                raise ValueError()
            major: int = int(match.group('major'))
            minor: int = int(match.group('minor'))
            patch: int = int(match.group('patch'))
            pre_release: PreRelease | None = None
            label_text: str | None = match.group('label')
            if label_text:
                pre_release = PreRelease(_MAP_OF_PRE_RELEASE_LABELS[label_text], int(match.group('number')))
        except (ValueError, TypeError, KeyError) as ex:
            raise VersionNumberError(f"'{text}' is not a Semantic Version number") from ex

        return VersionNumber(major, minor, patch, pre_release)

    @staticmethod
    def compare(left: IVersionNumber, right: IVersionNumber) -> int: # pylint: disable=too-many-return-statements, too-complex
        """
        (left == right) -> 0
        (left < right) -> -1
        (left > right) -> 1
        """
        if left is right:
            return 0
        if left.major < right.major:
            return -1
        if left.major > right.major:
            return 1
        if left.minor < right.minor:
            return -1
        if left.minor > right.minor:
            return 1
        if left.patch < right.patch:
            return -1
        if left.patch > right.patch:
            return 1
        if left.pre_release and not right.pre_release:
            return -1
        if not left.pre_release and right.pre_release:
            return 1
        if left.pre_release and right.pre_release:
            return PreRelease.compare(left.pre_release, right.pre_release)
        return 0

    def __str__(self) -> str:
        if self.pre_release:
            return f'{self.major}.{self.minor}.{self.patch}-{self.pre_release}'
        return f'{self.major}.{self.minor}.{self.patch}'

    def __hash__(self) -> int:
        return hash((self.major, self.minor, self.patch, hash(self.pre_release)))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, IVersionNumber):
            return VersionNumber.compare(self, other) == 0
        return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, IVersionNumber):
            return VersionNumber.compare(self, other) != 0
        return False

    def __lt__(self, other: IVersionNumber) -> bool:
        return VersionNumber.compare(self, other) < 0

    def __le__(self, other: IVersionNumber) -> bool:
        return VersionNumber.compare(self, other) <= 0

    def __gt__(self, other: IVersionNumber) -> bool:
        return VersionNumber.compare(self, other) > 0

    def __ge__(self, other: IVersionNumber) -> bool:
        return VersionNumber.compare(self, other) >= 0

_PRE_RELEASE_LABELS: list[str] = [label.value for label in PreReleaseLabel]
_MAP_OF_PRE_RELEASE_LABELS: dict[str, PreReleaseLabel] = {label.value: label for label in PreReleaseLabel}
_VERSION_NUMBER_PATTERN: re.Pattern[str] = re.compile(rf'(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<label>(?:{"|".join(_PRE_RELEASE_LABELS)}))\.(?P<number>[1-9]\d*))?')
