# MiloDB by FrozenWolf is marked with CC0 1.0.
# To view a copy of this license, visit http://creativecommons.org/publicdomain/zero/1.0
from abc import ABC, abstractmethod
from collections.abc import Sequence
from unittest import TestCase
from unittest.mock import Mock
from milodb_common.command_framework.argument_stream import ArgumentStream, ArgumentStreamError
from milodb_common.command_framework.candidate_text import CandidateText
from milodb_common.command_framework.i_command import ICommandLoader
from milodb_common.command_framework.i_command_constructor import ICommandConstructor
from milodb_common.parser.expanding_token_stream import ExpandingTokenStream
from milodb_common.parser.token import StartEnd, Token

_SET_OF_QUOTE_CHARS: set[str] = {'"', "'"}

def create_fake_tokens(list_of_token_text: list[str]) -> list[Token | None]:
    list_of_tokens: list[Token | None] = []
    token_number: int = 1
    token_start_index: int = 0
    for token_text in list_of_token_text:
        token_end_index: int = token_start_index + len(token_text)
        outer_indices: StartEnd = StartEnd(token_start_index, token_end_index)
        inner_indices: StartEnd = StartEnd(token_start_index, token_end_index)
        if token_text and token_text[0] in _SET_OF_QUOTE_CHARS:
            inner_indices.start += 1
            inner_indices.end -= 1
        list_of_tokens.append(Token(token_text, token_number, outer_indices, inner_indices, False))
        token_start_index = token_end_index + 1
        token_number += 1
    list_of_tokens.append(None)
    return list_of_tokens

class CommandTestBase(ABC, TestCase):
    def setUp(self) -> None:
        self.list_of_tokens: list[Token | None] = []
        self.mock_token_stream = Mock(ExpandingTokenStream)
        self.argument_stream: ArgumentStream | None = None

        self._command_loader: ICommandLoader | None = None
        self._load_exception: ArgumentStreamError | ICommandLoader.Error | ICommandConstructor.DelegateError | None = None
        self._load_result: ICommandLoader.Result | None = None

    @abstractmethod
    def create_command_loader(self) -> ICommandLoader:
        pass

    def command_load(self, list_of_token_text: list[str] | None, remaining_raw_text: str | None = None, expanded_raw_text: str | None = None) -> None:
        self._command_loader = self.create_command_loader()
        stub_command_token: Token | None = None
        self.list_of_tokens = [stub_command_token]
        if not list_of_token_text:
            self.mock_token_stream.next = Mock(return_value=None)
            self.mock_token_stream.get_remaining_raw_text = Mock(return_value='')
        else:
            self.list_of_tokens.extend(create_fake_tokens(list_of_token_text))
            self.mock_token_stream.next = Mock(side_effect=self.list_of_tokens[1:])
            self.mock_token_stream.get_expanded_raw_text = Mock(return_value=(expanded_raw_text or ''))
            self.mock_token_stream.get_remaining_raw_text = Mock(return_value=(remaining_raw_text or ''))
        self.argument_stream = ArgumentStream(self.mock_token_stream)
        try:
            self._load_result = self._command_loader.load(self.argument_stream)
        except (ArgumentStreamError, ICommandLoader.Error, ICommandConstructor.DelegateError) as ex:
            self._load_exception = ex

    def command_load_and_execute(self, list_of_token_text: list[str] | None, remaining_raw_text: str | None = None, expanded_raw_text: str | None = None) -> None:
        self.command_load(list_of_token_text, remaining_raw_text, expanded_raw_text)
        if self._load_exception is None:
            if self._load_result is not None:
                self._load_result.command.execute()
        else:
            self.fail(f"Load should have been successful but wasn't\nException was: {self._load_exception}")

    def assert_next_text(self, list_of_next_text: list[str]) -> None:
        self.assertIsNotNone(self._load_result)
        if self._load_result is not None:
            self._assert_candidate_text_list_equal(list_of_next_text, self._load_result.list_of_candidate_text)

    def assert_argument_error(self, fault_token_index: int | None, fault_text: str, list_of_candidate_text: Sequence[str | CandidateText]) -> None:
        self.assertIsInstance(self._load_exception, ArgumentStreamError)
        if isinstance(self._load_exception, ArgumentStreamError):
            if fault_token_index is not None:
                self.assertEqual(
                    self.list_of_tokens[fault_token_index], self._load_exception.fault_token,
                    'Unexpected fault token reported')
            else:
                self.assertIsNone(
                    self._load_exception.fault_token,
                    'Unexpected fault token reported')

            self.assertEqual(
                fault_text, self._load_exception.message,
                f"Expected error message of '{fault_text}' but got '{self._load_exception.message}'")

            self._assert_candidate_text_list_equal(list_of_candidate_text, self._load_exception.list_of_candidate_text)

    def assert_load_error(self, fault_token_index: int | None, fault_text: str, list_of_candidate_text: Sequence[str | CandidateText]) -> None:
        self.assertIsInstance(self._load_exception, ICommandLoader.Error)
        if isinstance(self._load_exception, ICommandLoader.Error):
            if fault_token_index is not None:
                self.assertEqual(
                    self.list_of_tokens[fault_token_index], self._load_exception.fault_token,
                    'Unexpected fault token reported')
            else:
                self.assertIsNone(
                    self._load_exception.fault_token,
                    'Unexpected fault token reported')

            self.assertEqual(
                fault_text, self._load_exception.message,
                f"Expected error message of '{fault_text}' but got '{self._load_exception.message}'")

            self._assert_candidate_text_list_equal(list_of_candidate_text, self._load_exception.list_of_candidate_text)

    def _assert_candidate_text_list_equal(self, expected_list_of_candidate_text: Sequence[str | CandidateText], actual_list_of_candidate_text: list[CandidateText]) -> None:
        expected_list_of_text: list[str] = []
        candidate_text: str | CandidateText
        for candidate_text in expected_list_of_candidate_text:
            if isinstance(candidate_text, CandidateText):
                expected_list_of_text.append(candidate_text.text + candidate_text.following_delimiter)
            else:
                expected_list_of_text.append(candidate_text + ' ')

        actual_list_of_text: list[str] = [candidate_text.text + candidate_text.following_delimiter for candidate_text in actual_list_of_candidate_text]

        self.assertListEqual(sorted(expected_list_of_text), sorted(actual_list_of_text))
